

/* dcf77.c
 * decoding of the dcf77-stream
 *
 * by: Clemens Helfmeier
 * This code converts the bitstream from the DCF77-sender into a standard
 * Timestamp
 */

#include "dcf77.h"
#include "string.h"

/* loacal variables */
/* the last correctly recieved timestamp. This will be given to the user
 * calling getTime(). If invalid, Year is set to 0;
 */
sTimeStamp tsLastValid;
// counts the number of times the timestamp is in/correctly recieved.
uint8_t u8ValidCounter = 0;
uint8_t u8InvalidCounter = 0;

/* the currently recieving timestamp */
sTimeStamp tsCurrent;

/* a very conservative timestamp that is taken to be veryvery correct */
sTimeStamp tsSecure;

uint8_t u8BitPosition;
uint8_t u8Status = 0;
uint8_t u8Value = 0;

#define Parity			(1<<0)
#define ParityError		(1<<1)
#define LastValid		(1<<2)
#define SecureValid		(1<<3)
#define CurrentValid	(1<<4)

#define ParityBit1	28
#define ParityBit2	35
#define ParityBit3	58
#define MonthBit	50
#define DayBit		42
#define SyncBit		59

/** get the current time
 * @return	Possible return values are:
 * 			0x00 on success,
 * 			0x02 no time available
 * 			0xFF invalid parameter
 */
extern uint8_t dcf77_getTime( pTimeStamp ts ) {
	if (ts == 0)
		return 0xFF;
	if ((u8Status & SecureValid) != 0) {
		/* there is a valid timestamp in the variable. Copy it! */
		memcpy( ts, &tsSecure, sizeof( tsSecure ) );
		return 0x00;
	}
	return 0x02;
}

/** Convert the given value from BCD (Binary Packed Decimal) to
 * normal binary (hex) numbers.
 */
uint8_t dcf77_bcd2hex( uint8_t bcd ) {
	return (bcd & 0x0F) + ((bcd >> 4) * 10);
}

/** Input one bit into the decoder
 * @param bit	zero means a bit of value 0, any other
 *				value is interpreted as a bit of value 1.
 */
void dcf77_putBit( uint8_t bit ) {
	/* put the bit into the parity flag */
	if (bit != 0) u8Status ^= Parity;
	/* don't save the bit in case we need 8 bit from the
	 * value later on. */
	/* handle the bit if there was something important */
	switch (u8BitPosition) {
		case 20:
			/* minute starts after here, clear the parity */
			u8Status &= ~(Parity);
			break;
		case ParityBit1:
			/* first parity bit */
			if ((u8Status & Parity) != 0)
				/* parity error in this word (minutes) */
				goto putBit_ParityError;
			/* parity was correct. */
			tsCurrent.minute = dcf77_bcd2hex((u8Value>>1) & 0x7F);
			break;
		case ParityBit2:
			/* second parity bit recieved */
			if ((u8Status & Parity) != 0)
				/* parity wasn't correct */
				goto putBit_ParityError;
			/* parity was correct */
			tsCurrent.hour = dcf77_bcd2hex((u8Value>>2) & 0x3F);
			break;
		case ParityBit3:
			// third parity bit recieved 
			if ((u8Status & Parity) != 0) {
				// parity wasn't correct
				goto putBit_ParityError;
			}
			// parity was correct 
			tsCurrent.year = dcf77_bcd2hex((u8Value) & 0xFF) + 2000;
			break;
		case MonthBit:
			// the last bit of the month was recieved
			tsCurrent.month = dcf77_bcd2hex((u8Value>>3) & 0x1F);
			break;
		case DayBit:
			// the last bit was the last day bit
			tsCurrent.day = dcf77_bcd2hex((u8Value>>2) & 0x3F);
			break;
		case 4:
			// if the current timestamp and the last valid timestamp
			// are both valid, compare them:
			if ((u8Status & (CurrentValid | LastValid)) == (CurrentValid | LastValid)) {
				// are both timestamp equal?
				if ((tsLastValid.year == tsCurrent.year) &&
								 (tsLastValid.month == tsCurrent.month) &&
								 (tsLastValid.day == tsCurrent.day) &&
								 (tsLastValid.hour == tsCurrent.hour) &&
								 (tsLastValid.minute == tsCurrent.minute))
					u8ValidCounter++;
				else
					u8InvalidCounter++;
				// did we already verify the timestamp enough?
				if (u8ValidCounter + u8InvalidCounter >= 3) {
					if (u8ValidCounter > u8InvalidCounter) {
					// the Timestamp is taken to be correct
					// verify the timestamp again.
						u8ValidCounter = 0;
						u8InvalidCounter = 0;
						memcpy( &tsSecure, &tsLastValid, sizeof( tsSecure ) );
						u8Status |= SecureValid;
					} else {
					// there are more different timestamps recieved, discard the current one
						u8Status &= ~LastValid;
					}
				}
			}
			// invalidate the current timestamp:
			u8Status &= ~CurrentValid;
			break;
	} /* switch */
	/* now put the new bit into the queue */
	if (bit == 0)
		u8Value = (u8Value >> 1);
	else {
		u8Value = (u8Value >> 1) | 0x80;
	}
	/* increase the bit position */
	u8BitPosition++;
	return;
putBit_ParityError:
		// reset the parity bit
		u8Status &= ~Parity;
		// remember the error
		u8Status |= ParityError;
}

/** input the synchronizeation bit into the decoder
 */
void dcf77_putSync( void ) {
	if ((u8BitPosition == SyncBit) && ((u8Status & ParityError) == 0)) {
		// the sync bit is at the correct position, there was no error
		u8Status |= CurrentValid;
		if ((u8Status & LastValid) == 0) {
			// this is the first timestamp recieved, copy it
			memcpy( &tsLastValid, &tsCurrent, sizeof( tsLastValid ) );
			tsLastValid.second = 0;
			u8ValidCounter = 0;
			u8InvalidCounter = 0;
			u8Status ^= LastValid | CurrentValid;	// remove the current valid, we don't want to compare them already, they are equal.
		}
	}
	/* reset Parity and Error bits */
	u8Status &= ~(Parity | ParityError);
	u8BitPosition = 0;	/* synchronize */
	return;
}

/** increase the given timestamp by one second.
 */
void dcf77_incTS( sTimeStamp *ts ) {
	if (++ts->second >= 60) {
		ts->second = 0;
		if (++ts->minute >= 60) {
			ts->minute = 0;
			if (++ts->hour >= 24) {
				ts->hour = 0;
				ts->day++;
				uint8_t maxday = 30;
				if (ts->month == 2)
					maxday=28;
				else if ((ts->month < 8) && ((ts->month & 1) != 0))
					maxday = 31;
				else if ((ts->month >= 8) && ((ts->month & 1) == 0))
					maxday = 31;
				if (ts->day > maxday) {
					ts->day = 1;
					if (++ts->month > 12) {
						ts->month = 1;
						ts->year++;
					}
				}
			}
		}
	}
}
/** Should be called regularily with exactly 1 Hz. It keeps track of the time.
 */

void dcf77_timer( void ) {
	if ((u8Status & LastValid) != 0) {
		dcf77_incTS( &tsLastValid );
	}
	if ((u8Status & SecureValid) != 0) {
		dcf77_incTS( &tsSecure );
	}
}




